/* rpcfunc.c */

/*
 *  Copyright (c) 1992, Research Systems Inc.  All rights reserved.
 *   Reproduction by any means whatsoever is  prohibited without express
 *   written permission.
 */

/************************************************************************/
/*									*/
/*	User library routines to allow C-Routine to be RPC client	*/
/*	to IDL running as a server					*/
/*									*/
/************************************************************************/


#include        <stdio.h>
#include        <ctype.h>
#include        <rpc/rpc.h>

#ifdef MIPS
#include	<bsd/sys/time.h>
#else
#include	<sys/time.h>
#endif
#include	<sys/socket.h>
#include	<netdb.h>
#include	"export.h"
#include	"rpc_idl.h"



/************************************************************************/
/*	Default time out is 1 minute but can be overridden by user	*/
/************************************************************************/

static struct timeval	time_out	= { 60, 0 };
static int		verbose		= 0;


/************************************************************************/
/*									*/
/*	register_idl_client						*/
/*									*/
/*	set up a connection with the specified server. Any of the	*/
/*	3 input parameters will have the default value used if		*/
/*	the value is 0 (or NULL).					*/
/*									*/
/************************************************************************/


CLIENT	*register_idl_client( server_id, hostname, timeout )
    LONG		server_id;
    char		*hostname;
    struct timeval	*timeout;
{
    CLIENT		*client;
    int			sock		= RPC_ANYSOCK;
    struct hostent	*hp;
    struct sockaddr_in	server_addr;

#define		BUF_SIZ		(8*1024)

    /*	Use defaults is user hasn't provided values			*/

    if( server_id == 0 )
	server_id	= IDL_DEFAULT_ID;

    if( hostname == NULL || *hostname == '\0' )
	hostname	= "localhost";

    (void)set_idl_timeout( timeout );

    /*	Create a TCP/IP link to the server				*/

    /*	Unknown host?							*/

    if( (hp = gethostbyname( hostname )) == NULL ) {
	/* error message here */
	return NULL;
    }

    bcopy( hp->h_addr, (caddr_t)&server_addr.sin_addr, hp->h_length );
    server_addr.sin_family	= AF_INET;
    server_addr.sin_port	= 0;

    if(( client = clnttcp_create(&server_addr,
				 server_id,
				 IDL_DEFAULT_VERSION,
				 &sock, BUF_SIZ, BUF_SIZ )) == NULL ) {
	if( verbose )
	    clnt_pcreateerror("clnttcp_create");
    }
    return client;
}

/************************************************************************/
/*									*/
/*	unregister_idl_client						*/
/*									*/
/*	elegant clean up of client data					*/
/*									*/
/************************************************************************/

void	unregister_idl_client( client )
    CLIENT	*client;
{
    if( client != NULL )
	clnt_destroy( client );
}

/************************************************************************/
/*									*/
/*	kill_server							*/
/*									*/
/*	ask the server to exit.						*/
/*									*/
/************************************************************************/

int	kill_server( client )
    CLIENT	*client;
{
    enum clnt_stat	result;

    result	= clnt_call(client, FORCED_EXIT,
			    xdr_void, 0, xdr_void, 0,
			    time_out);
    if( result != RPC_SUCCESS ) {
	if(verbose)
	    clnt_perror( client, "KILL_SERVER" );
	return 0;
    }
    return 1;
}


/************************************************************************/
/*									*/
/*	set_idl_timeout							*/
/*									*/
/*	Sets RPC timeout						*/
/*	Returns TRUE if time_out is set.				*/
/*									*/
/************************************************************************/

int	set_idl_timeout( timeout )
    struct timeval	*timeout;
{
    if( timeout != NULL && timerisset( timeout ) ) {
	time_out.tv_sec 	= timeout->tv_sec;
	time_out.tv_usec	= timeout->tv_usec;
	return 1;
    }
    return 0;
}

/************************************************************************/
/*									*/
/*	set_rpc_verbosity						*/
/*									*/
/*	If verbose == TRUE. Don't write any error messages.		*/
/*	If verbose == FALSE. Write error messages.			*/
/************************************************************************/

void	set_rpc_verbosity( user_verbose )
    int		user_verbose;
{
    verbose	= (user_verbose != 0);
}


/************************************************************************/
/*									*/
/*	send_idl_command						*/
/*									*/
/*	send a command to the the server. returns 1 if successful,	*/
/*	and the IDL error message if not (a small negative integer)	*/
/*									*/
/*	The command can be anything one would type at the prompt.	*/
/*	E.g. print,'Hello World' & shade_surf,dist(20)			*/
/*									*/
/************************************************************************/

int	send_idl_command( client, command )
    CLIENT	* client;
    char	* command;
{
    int			result;
    int			comm_result;

    /*	Don't send NULL commands: wrapstring barfs			*/

    if( command == NULL )
	return(-2);

    comm_result	= clnt_call( client, RUN_COMMAND,
		       xdr_wrapstring, &command,
		       xdr_int, &result,
		       time_out );

    /* comm failure */

    if( comm_result != RPC_SUCCESS ) {
	if(verbose)
	    clnt_perror( client, "RUN_COMMAND" );
	return(-1);
    }

    return result;
}

/************************************************************************/
/*									*/
/*	idl_server_interactive						*/
/*									*/
/*	make server go interactive					*/
/*									*/
/************************************************************************/

int	idl_server_interactive( client )
    CLIENT	*client;
{
    enum clnt_stat	result;

    result	= clnt_call(client, RUN_INTERACTIVE,
			    xdr_void, 0, xdr_void, 0,
			    time_out);
    if( result != RPC_SUCCESS ) {
	/*	It is very likely we will timeout in this mode		*/
	/*	We will only complain if the result is not a timeout	*/
	
	if(result != RPC_TIMEDOUT && verbose)
	    clnt_perror( client, "RUN_INTERACTIVE" );
	return 0;
    }
    return 1;
}


/************************************************************************/
/*									*/
/*	var_ok								*/
/*									*/
/*	Sanity check on varinfo_t contents. 				*/
/*									*/
/************************************************************************/

static int	var_ok( var )
    varinfo_t	* var;
{
    VARIABLE	* v;
    REGISTER	char	*name, ch;

    /*	Check that the name is valid					*/

    name	= var->Name;
    ch		= *name++;
    if( !isalpha(ch) && ch != '_' )
	return 0;
    while( ch = *name++ ) {
	if( !isalpha(ch) && !isdigit(ch) && ch != '$' && ch != '_' )
	    return 0;
    }

    /*	Now check the variable						*/

    if( (v=var->Variable) == NULL )
	return 0;

    if( v->type > MAX_SIMPLE_TYPE )
	return 0;

    /*	Only allow valid flags through					*/

    v->flags	&= V_ARR;

    /*	Make sure to add V_DYNAMIC as needed.				*/

    if( v->type == TYP_STRING || (v->flags & V_ARR) )
	v->flags	|= V_DYNAMIC;

    /*	Try to make sure that array variables have actual data		*/

    if( v->flags & V_ARR ) {
	if( v->value.arr->data == NULL )
	    return 0;
    }

    /*	Probably OK							*/

    return 1;
}


/************************************************************************/
/*									*/
/*	set_idl_variable						*/
/*									*/
/*	Send a variable to IDL to put into $MAIN.			*/
/*	Return 1 for success, 0 for communication failure,		*/
/*	-1 if variable is malformed.					*/
/*									*/
/************************************************************************/

int	set_idl_variable( client, var )
    CLIENT	* client;
    varinfo_t	* var;
{
    int			result;
    enum clnt_stat	comm_result;

    if( var_ok( var ) ) {
	comm_result	= clnt_call( client, SET_VARIABLE,
				    xdr_vinfo, var,
				    xdr_void, 0,
				    time_out );

	/* comm failure */

	if( comm_result != RPC_SUCCESS ) {
	    if(verbose)
		clnt_perror( client, "SET_VARIABLE" );
	    return 0;
	}
	return 1;
    }
    return -1;		/* malformed variable	*/
}

/************************************************************************/
/*									*/
/*	get_idl_variable						*/
/*									*/
/*	Read the value of a variable out of IDL's $MAIN procedure level	*/
/*	Return 1 for success, 0 for communication failure,		*/
/*	-1 if arguments are bad.					*/
/*	-2 if the name is illegal (E.g. 2138  #abc  !DEVICE)		*/
/*	-3 variable is not transportable (structure, ASSOC'ed var)	*/
/*									*/
/************************************************************************/

int	get_idl_variable( client, name, var, typecode )
    CLIENT	* client;
    char	* name;
    varinfo_t	* var;
    int		typecode;
{
    int			result;
    int			comm_result;
    /*	Make this static to ensure that the MAXIDLEN'th character of	*/
    /*	Nameis always NULL						*/
    static getinfo_t	getinfo = { 0 };

    if( name == NULL || var == NULL )
	return -1;

    strncpy(getinfo.Name, name, MAXIDLEN);
    getinfo.TypeCode	= typecode;
    comm_result	= clnt_call( client, GET_VARIABLE,
		       xdr_idl_getinfo, &getinfo,
		       xdr_vinfo, var,
		       time_out );

    /* comm failure */

    if( comm_result != RPC_SUCCESS ) {
	if(verbose)
	    clnt_perror( client, "GET_VARIABLE" );
	return( 0 );
    }

    /*	If IDL was unhappy with the variable (bad naming, or 		*/
    /*	an untransmittable variable the name returned will be		*/
    /*	<ILLEGAL-NAME> or <BAD-VAR-TYPE>				*/

    if( var->Name[0] == '<' ) {
	if( var->Name[1] == 'I' )
	    return -2;
	return -3;
    }
    
    return 1;
}


/************************************************************************/
/*									*/
/*	set_idl_shm_variable						*/
/*									*/
/*	Send a variable to IDL to put into $MAIN.			*/
/*	Return 1 for success, 0 for communication failure,		*/
/*	-1 if variable is malformed.					*/
/*									*/
/************************************************************************/

int	set_idl_shm_variable( client, vinfo, id, offset, shm_info )
    CLIENT	* client;
    varinfo_t	* vinfo;
    int		id;
    LONG	offset;
    shminfo_t	*shm_info;
{
    int			result;
    enum clnt_stat	comm_result;

    /*	Only send arrays via shared memory			*/
    if( var_ok( vinfo ) ) {
	if((vinfo->Variable->flags & V_ARR) == 0)
	    return set_idl_variable(client, vinfo);

	/*	Copy data and set up structure to point at the	*/
	/*	substructure within the shared memory space	*/
	
	shm_info->shm_data_id		= id;
	shm_info->shm_data_offset	= offset;
	shm_info->shm_vinfo		= *vinfo;
	shm_info->u.v.shm_var		= *vinfo->Variable;
	shm_info->u.v.shm_array		= *vinfo->Variable->value.arr;
	
	comm_result	= clnt_call( client, SHM_SET_VARIABLE,
				    xdr_void, 0,
				    xdr_void, 0,
				    time_out );

	/* comm failure */

	if( comm_result != RPC_SUCCESS ) {
	    if(verbose)
		clnt_perror( client, "SET_SHM_VARIABLE" );
	    return 0;
	}
	return 1;
    }
    return -1;		/* malformed variable	*/
}
